"""
.. module:: is_bad_dq
   :synopsis: Returns True if a given DQ (or array of DQ values) fail(s) a test
       defined in this module, otherwise returns False.  If an array of DQ
       values was given, then an array of True/False values is returned for each
       element in the array, otherwise a single True/False value is returned.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
#--------------------
# External Imports
#--------------------
import numpy
#--------------------
# Package Imports
#--------------------
from spec_plots import __version__
#--------------------

def is_bad_dq(instrument, dqs):
    """
    Returns True/False whether the DQ values are from a good part of the
    spectrum.  If the input DQs is a scalar value, the return result is also
    a scalar value.  If the input DQs are a numpy array, the return result
    is also a numpy array.

    :param instrument: The instrument the DQs come from.

    :type instrument: str

    :param dqs: Array of DQ (STIS) or DQ_WGT (COS) values.

    :type dqs: numpy.ndarray

    :returns: bool or numpy.ndarray -- A scalar boolean or array of booleans.
    """

    if instrument == "cos":
        if isinstance(dqs, numpy.ndarray):
            return numpy.asarray([x < 1 for x in dqs])
        else:
            return dqs < 1

    elif instrument == "stis":
        if isinstance(dqs, numpy.ndarray):
            return numpy.asarray([x != 0 and x != 16 for x in dqs])
        else:
            return dqs != 0 and dqs != 16

    elif instrument in ["miri", "nirspec", "niriss"]:
        if isinstance(dqs, numpy.ndarray):
            return numpy.asarray([x < 1 for x in dqs])
        else:
            return dqs < 1

    else:
        return numpy.asarray([])
#--------------------



"""
.. module:: plotspec
   :synopsis: Creates preview plots for the provided JWST spectrum.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
from __future__ import division
import copy
import os
import sys
from builtins import str
#--------------------
# External Imports
#--------------------
import matplotlib
from matplotlib.ticker import FormatStrFormatter
import matplotlib.pyplot as pyplot
from matplotlib import rc
import numpy
#--------------------
# Package Imports
#--------------------
from spec_plots.utils.specutils.specutilserror import SpecUtilsError
from spec_plots.utils.specutils.debug_oplot import debug_oplot
from spec_plots.utils.specutils.calc_covering_fraction import (
    calc_covering_fraction)
from spec_plots import __version__

if matplotlib.get_backend().lower() != 'agg':
    pyplot.switch_backend('Agg')
#--------------------

#--------------------
def plotspec(jwst_spectrum, output_type, output_file, flux_scale_factor,
             fluxerr_scale_factor, plot_metrics, dpi_val=96., output_size=1024,
             debug=False, full_ylabels=False, optimize=True):
    """
    Accepts a JWSTSpectrum object from the READSPEC function and produces
    preview plots.

    :param jwst_spectrum: JWST spectrum as returned by READSPEC.

    :type jwst_spectrum: JWSTSpectrum

    :param output_type: What kind of output to make?

    :type output_type: str

    :param output_file: Name of output file (including full path).

    :type output_file: str

    :param flux_scale_factor: Max. allowed ratio between the flux and a median
        flux value, used in edge trimming.  Default = 10.

    :type flux_scale_factor: float

    :param fluxerr_scale_factor: Max. allowed ratio between the flux uncertainty
        and a median flux uncertainty value, used in edge trimming.
        Default = 5.

    :type fluxerr_scale_factor: float

    :param plot_metrics: Collection of plot metrics (flux statistics, axis
        ranges, etc.) to use when making the plots.  These are computed using
        `utils.specutils.calc_plot_metrics()`.

    :type plot_metrics: dict

    :param dpi_val: The DPI value of your device's monitor.  Affects the size of
        the output plots.  Default = 96. (applicable to most modern monitors).

    :type dpi_val: float

    :param output_size: Size of plot in pixels (plots are square in dimensions).
        Defaults to 1024.

    :param output_size: int

    :param debug: Should the output plots include debugging information
        (color-coded data points based on rejection criteria, shaded exclude
        regions)?  Default = False.

    :type debug: bool

    :param full_ylabels: Should the y-labels contain the full values (including
        the power of 10 in scientific notation)?  Default = False.

    :type full_ylabels: bool

    :param optimize: If set to True, will use a slightly optimized version of
        determining the plot covering fraction.

    :type optimize: bool

    :raises: OSError, utils.specutils.SpecUtilsError

    .. note::

         This function assumes a screen resolution of 96 DPI in order to
         generate plots of the desired sizes.  This is because matplotlib works
         in units of inches and DPI rather than pixels.
    """

    # Make sure the plot size is set to an integer value.
    if output_size is not None:
        if not isinstance(output_size, int):
            output_size = int(round(output_size))

    # Make sure the output path exists, if not, create it.
    if output_type != 'screen':
        if (os.path.dirname(output_file) != "" and
                not os.path.isdir(os.path.dirname(output_file))):
            try:
                os.mkdir(os.path.dirname(output_file))
            except OSError as this_error:
                if this_error.errno == 13:
                    sys.stderr.write("*** MAKE_JWST_SPEC_PREVIEWS ERROR:"
                                     " Output directory could not be created,"
                                     " "+repr(this_error.strerror)+"\n")
                    exit(1)
                else:
                    raise

    is_bigplot = output_size > 128

    # Start plot figure.
    this_figure, this_plotarea = pyplot.subplots(nrows=1, ncols=1,
                                                 figsize=(
                                                     output_size/dpi_val,
                                                     output_size/dpi_val),
                                                 dpi=dpi_val)

    # Adjust the plot geometry (margins, etc.) based on plot size.
    if is_bigplot:
        this_figure.subplots_adjust(hspace=0.3, top=0.915)
    else:
        this_figure.subplots_adjust(top=0.85, bottom=0.3, left=0.25, right=0.8)

    # Get the wavelengths, fluxes, flux uncertainties, and data quality
    # flags out of the spectrum.
    try:
        all_wls = jwst_spectrum.wavelengths
        all_fls = jwst_spectrum.fluxes
        all_flerrs = jwst_spectrum.fluxerrs
        all_dqs = jwst_spectrum.dqs
    except KeyError as the_error:
        raise SpecUtilsError("The provided stitched spectrum does not have"
                             " the expected format, missing key " +
                             str(the_error)+".")

    # Extract the optimal x-axis plot range from the plot_metrics dict,
    # since we use it a lot.
    optimal_xaxis_range = plot_metrics["optimal_xaxis_range"]

    # Plot the spectrum, but only if valid wavelength ranges for x-axis
    # are returned, otherwise plot a special "Fluxes Are All Zero" plot.
    if all(numpy.isfinite(optimal_xaxis_range)):
        # We plot the spectrum as a regular line for use
        # in calc_covering_fraction, it will be removed later.
        this_line = this_plotarea.plot(all_wls, all_fls, 'b')

        # Update y-axis range, but only adjust the ranges if this isn't
        # an all-zero flux case (and not in debug mode, in which case I want
        # to see the entire y-axis range).
        if not debug:
            this_plotarea.set_ylim(plot_metrics["y_axis_range"])

        covering_fractions = calc_covering_fraction(
            this_figure, numpy.asarray([this_plotarea]), 0, optimize=optimize)
        # Note: here we remove the line we plotted before, it was only
        # so that calc_covering_fraction would have someting to draw on the
        # canvas and thereby determine which pixels were "blue" (i.e., part
        # of the plotted spectrum vs. background).
        this_line.pop(0).remove()
        # Now we plot the spectrum as a LineCollection so that the
        # transparency will have the desired effect, but, this is not
        # rendered on the canvas inside calc_covering_fraction, hence why we
        # need to plot it both as a regular line first.
        # Note: because we are re-using a LineCollection object in the
        # array of plot_metrics (specifically, when creating the
        # thumbnail-sized plot), we have to use a copy of the LineCollection
        # object, otherwise it will have Axes, Figure, etc. all defined and
        # resetting them to None does not work.  Since this is only an issue
        # with thumbnail-sizes, this is only relevant for the first
        # LineCollection.
        this_collection = this_plotarea.add_collection(copy.copy(
            plot_metrics["line_collection"]))

        if covering_fractions > 30.:
            this_collection.set_alpha(0.1)

        # Turn on plot grid lines.
        this_plotarea.grid(True, linestyle="dashed")

        if is_bigplot:
            this_figure.suptitle(os.path.basename(
                jwst_spectrum.orig_file), fontsize=18, color='r')

        if debug:
            # Overplot points color-coded based on rejection criteria.
            debug_oplot(this_plotarea, "jwst", all_wls, all_fls,
                        all_flerrs, all_dqs,
                        plot_metrics["median_flux"],
                        plot_metrics["median_fluxerr"],
                        flux_scale_factor, fluxerr_scale_factor,
                        plot_metrics["fluxerr_95th"])

            # Overplot the x-axis edges that are trimmed to define the
            # y-axis plot range as a shaded area.
            this_plotarea.axvspan(numpy.nanmin(all_wls),
                                  optimal_xaxis_range[0],
                                  facecolor="lightgrey", alpha=0.5)
            this_plotarea.axvspan(optimal_xaxis_range[1],
                                  numpy.nanmax(all_wls),
                                  facecolor="lightgrey", alpha=0.5)

            # Overplot the avoid regions in a light color as a shaded area.
            for region in plot_metrics["avoid_regions"]:
                this_plotarea.axvspan(region.minwl, region.maxwl,
                                      facecolor="lightgrey", alpha=0.5)

        # This is where we ensure the x-axis range is based on the full
        # x-axis range, rather than using the optimum x-axis range.  This is
        # done so that all the plots for a similar instrument setting will
        # have the same starting and ending plot values.
        min_wl = numpy.nanmin(all_wls)
        max_wl = numpy.nanmax(all_wls)
        xplot_buffer = (max_wl - min_wl) * 0.05
        this_plotarea.set_xlim([min_wl-xplot_buffer, max_wl+xplot_buffer])

        # Only use two tick labels (min and max wavelengths) for
        # thumbnails, because there isn't enough space otherwise.
        if not is_bigplot:
            rc('font', size=10)
            minwl = numpy.nanmin(all_wls)
            maxwl = numpy.nanmax(all_wls)
            this_plotarea.set_xticks([minwl, maxwl])
            this_plotarea.set_xticklabels(this_plotarea.get_xticks(),
                                          rotation=25.)
            this_plotarea.xaxis.set_major_formatter(FormatStrFormatter(
                "%6.1f"))
        else:
            # Make sure the font properties go back to normal.
            pyplot.rcdefaults()
            this_plotarea.tick_params(axis='x', labelsize=14)
            this_plotarea.set_xlabel(r"Wavelength $(\AA)$", fontsize=16,
                                     color='k')
            this_plotarea.set_ylabel(r"Flux $\mathrm{(erg/s/cm^2\!/\AA)}$",
                                     fontsize=16, color='k')

            # If requested, include the powers of 10 part of the y-axis
            # tickmarks.
            if full_ylabels:
                this_plotarea.yaxis.set_major_formatter(FormatStrFormatter(
                    '%3.2E'))

    else:
        # Otherwise this is a spectrum that has all zero fluxes, or some
        # other problem, and we make a default plot.  Define the optimal
        # x-axis range to span the original spectrum.
        optimal_xaxis_range = [numpy.nanmin(all_wls), numpy.nanmax(all_wls)]
        this_plotarea.set_xlim(optimal_xaxis_range)

        # Make the plot background grey to distinguish that this is a
        # `special` plot.  Turn off y-tick labels.
        this_plotarea.set_facecolor("lightgrey")
        this_plotarea.set_yticklabels([])

        # Configure the plot units, text size, and other markings based
        # on whether this is a large or thumbnail-sized plot.
        if not is_bigplot:
            rc('font', size=10)
            minwl = numpy.nanmin(all_wls)
            maxwl = numpy.nanmax(all_wls)
            this_plotarea.set_xticks([minwl, maxwl])
            this_plotarea.set_xticklabels(this_plotarea.get_xticks(),
                                          rotation=25.)
            this_plotarea.xaxis.set_major_formatter(FormatStrFormatter(
                "%6.1f"))
            textsize = "small"
            plottext = "Fluxes are \n all 0."
        else:
            # Make sure the font properties go back to normal.
            pyplot.rcdefaults()
            this_plotarea.tick_params(axis='x', labelsize=14)
            this_plotarea.set_xlabel(r"Wavelength $(\AA)$", fontsize=16,
                                     color='k')
            this_plotarea.set_ylabel(r"Flux $\mathrm{(erg/s/cm^2\!/\AA)}$",
                                     fontsize=16, color='k')

            # If requested, include the powers of 10 part of the y-axis
            # tickmarks.
            if full_ylabels:
                this_plotarea.yaxis.set_major_formatter(FormatStrFormatter(
                    '%3.2E'))

            textsize = "x-large"
            plottext = "Fluxes are all 0."

        # Place the text with the informational message in the center of
        # the plot.
        this_plotarea.text(0.5, 0.5, plottext, horizontalalignment="center",
                           verticalalignment="center",
                           transform=this_plotarea.transAxes, size=textsize)

    # Display or plot to the desired format.
    if output_type != "screen":
        if output_size == 128:
            revised_output_file = output_file.strip('\.png') + '_thumb.png'
        else:
            revised_output_file = output_file

        # Save figure.
        this_figure.savefig(revised_output_file, format=output_type,
                            dpi=dpi_val)

    elif output_type == "screen":
        pyplot.show()
#--------------------



"""
.. module:: __main_jwst__

   :synopsis: This script is used when make_jwst_spec_previews is called
   directly as an executable from the command line.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
import sys
#--------------------
# Package Imports
#--------------------
import spec_plots.make_jwst_spec_previews as mjwsp
from spec_plots import __version__

#--------------------

def main():
    """ This is the main function used when invoking make_jwst_spec_previews
    directly as an executable. """

    # Collect command-line arguments and package into an ArgParse object.
    parser = mjwsp.setup_args()
    args = parser.parse_args(sys.argv[1:])

    # Check arguments and options.
    mjwsp.check_input_options(args)

    # Call main function.
    mjwsp.make_jwst_spec_previews(args.input_file,
                                  flux_scale_factor=args.flux_scale_factor,
                                  fluxerr_scale_factor=
                                  args.fluxerr_scale_factor,
                                  n_consecutive=args.n_consecutive,
                                  output_path=args.output_path,
                                  output_type=args.output_type,
                                  dpi_val=args.dpi_val,
                                  debug=args.debug,
                                  full_ylabels=args.full_ylabels,
                                  optimize=not args.nooptimize,
                                  verbose=args.verbose)

if __name__ == "__main__":
    main()



"""
.. module:: get_flux_stats
   :synopsis: Calculates statistics on an arrays of fluxes and corresponding
       flux uncertainties, such as the median values and the 95th percentile
       values.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
#--------------------
# External Imports
#--------------------
import numpy
#--------------------
# Package Imports
#--------------------
from spec_plots import __version__

#--------------------

def get_flux_stats(fluxes, fluxerrs):
    """
    Calculates median values of the fluxes and flux uncertainties, and the 95th
    percentile flux uncertainty.

    :param fluxes: The fluxes to be plotted.

    :type fluxes: numpy.ndarray

    :param fluxerrs: The uncertainties of the fluxes to be plotted.

    :type fluxerrs: numpy.ndarray

    :returns: tuple -- The median flux, median flux uncertainty, and the 95th
        percentile flux uncertainty.
    """

    # Find the median flux value, ignoring any NaN values or fluxes that are
    # 0.0.
    where_finite_and_notzero = numpy.where((numpy.isfinite(fluxes)) &
                                           (fluxes != 0.0))

    if len(where_finite_and_notzero[0]) > 0:
        median_flux = numpy.median(fluxes[where_finite_and_notzero])
        median_fluxerr = numpy.median(fluxerrs[where_finite_and_notzero])
    else:
        median_flux = numpy.nan
        median_fluxerr = numpy.nan

    # Get the 95th percentile flux uncertainty value.
    fluxerr_95th = numpy.percentile(fluxerrs, 95.)

    return median_flux, median_fluxerr, fluxerr_95th
#--------------------



"""
.. module:: calc_covering_fraction
   :synopsis: Calculates the ratio of the number of pixels containing plot data
       vs. empty (background) plot pixels.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
from __future__ import division
from builtins import range
#--------------------
# External Imports
#--------------------
import numpy
#--------------------
# Package Imports
#--------------------
from spec_plots import __version__

#--------------------

def count_blue_red(buf):
    """
    Counts the number of "blue" vs. "red" pixels in the plot area.

    :param buf: The plot pixels read into a buffer of RGB values (
        one-dimensional, [r,g,b,r,g,b,...]).

    :type buf: numpy.ndarray

    :returns: tuple -- The number of blue and red pixels, respectively.  Blue
        represent pixels that have plot data in them, red represent background
        pixels that do not have any plot lines going through them.
    """

    blue_count = 0
    red_count = 0
    for i in range(0, len(buf), 3):
        # First make sure this isn't a grey pixel, where R=G=B.
        if not (buf[i] == buf[i+1] and buf[i] == buf[i+2]):
            if buf[i] > 0 and buf[i+2] == 0:
                red_count += 1
            elif buf[i+2] > 0:
                blue_count += 1

    return blue_count, red_count
#--------------------

#--------------------
def calc_covering_fraction(fig, subplots, subplot_num, optimize=True):
    """
    Calculates the ratio of pixels that contain plot data vs. background (empty)
    pixels by temporarily filling the background with RGB value = red and
    then identifying blue (plot data) pixels.  This ratio is used to
    determine whether to make the plot lines transparent or not for readability.

    :param fig: The plot figure.

    :type fig: Figure

    :param subplots: The list of subplots.

    :type subplots: list

    :param subplot_num: The index indicating which subplot is currently being
        examined.

    :type subplot_num: int

    :param optimize: If set to True, will use a slightly optimized version of
        determining the plot covering fraction.  This has been tested to
        *almost* always give the same result as the non-optimized version.
        There has been (rare) instances where the covering fractions differed
        slightly, so if the preference is to have more accurate/robust covering
        fractions, set optimize=False!  The speed benefit is on the order of
        10-20%, but is largest for plots with multiple subplots.

    :type optimize: bool

    :returns: float -- The ratio of pixels containing plot data ("blue") vs.
        empty ("red") pixels, as a percentage (0. <= p <= 100.).
    """

    # Turn off all other subplots since the RGB array is for the entire
    # figure, not just this subplot part of it.
    for i, subplot in enumerate(subplots):
        if i != subplot_num:
            subplot.set_visible(False)
        else:
            subplot.set_visible(True)

    # Temporarily make the background color red so that it's easy to identify
    # empty pixels.
    subplots[subplot_num].set_facecolor('red')

    # Draw the plot, convert into numpy array of RGB values.
    fig.canvas.draw()
    buf = numpy.fromstring(fig.canvas.tostring_rgb(), dtype=numpy.uint8)

    # Count the number of "blue" and "red" pixels.  The optimized version
    # attempts to split the RGB array into equal slices so that the counting of
    # red and blue pixels can be done quicker.  Early tests have shown this
    # usually produces the same results.
    if optimize:
        n_per_subplot = int(numpy.ceil(len(buf) / len(subplots)))
        while n_per_subplot % 3 != 0:
            n_per_subplot += 1
        startindex = subplot_num * n_per_subplot
        endindex = startindex + n_per_subplot
        if subplot_num < len(subplots)-1:
            blue_count, red_count = count_blue_red(buf[startindex:endindex])
        else:
            blue_count, red_count = count_blue_red(buf[startindex:])
    else:
        blue_count, red_count = count_blue_red(buf)

    # If this is the last subplot, then make sure we make all the other
    # subplots visible again.
    if subplot_num == subplots.shape[0]-1:
        for subplot in subplots:
            subplot.set_visible(True)

    # Make the background color white again for this subplot.
    subplots[subplot_num].set_facecolor('white')

    return float(blue_count) / float(red_count+blue_count) * 100.
#--------------------



"""
.. module:: make_fits
   :synopsis: Makes a binary FITS table out of the provided COS spectrum.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
from __future__ import division
import os
import sys
#--------------------
# External Imports
#--------------------
from astropy.io import fits
import numpy
#--------------------
# Package Imports
#--------------------
from spec_plots.utils.specutils_cos.get_segment_names import get_segment_names
from spec_plots import __version__
#--------------------

#--------------------
def make_fits(cos_spectrum, output_file, plot_metrics, input_file):
    """
    Accepts a COSSpectrum object from the READSPEC function and produces a
    binary FITS table.

    :param cos_spectrum: COS spectrum as returned by READSPEC.

    :type cos_spectrum: COSSpectrum

    :param output_file: Name of output file (including full path).

    :type output_file: str

    :param plot_metrics: Collection of plot metrics (flux statistics, axis
        ranges, etc.) to use when making the plots.  These are computed using
        `utils.specutils.calc_plot_metrics()`.

    :type plot_metrics: list

    :param input_file: The full path and name of the input file we are making
        a preview spectrum for.

    :type input_file: str
    """

    # Make sure the output path exists, if not, create it.
    if (os.path.dirname(output_file) != "" and
            not os.path.isdir(os.path.dirname(output_file))):
        try:
            os.mkdir(os.path.dirname(output_file))
        except OSError as this_error:
            if this_error.errno == 13:
                sys.stderr.write("*** MAKE_HST_SPEC_PREVIEWS ERROR: Output"
                                 " directory could not be created, " +
                                 repr(this_error.strerror)+"\n")
                exit(1)
            else:
                raise

    # Get list of segment names.
    segment_names = get_segment_names(cos_spectrum)

    # Loop over each segment.
    segment_ylims = []
    for i, segment in enumerate(segment_names):
        if i == 0:
            all_wls = cos_spectrum.segments[segment].wavelengths
            n_wls = len(cos_spectrum.segments[segment].wavelengths)
            all_fls = cos_spectrum.segments[segment].fluxes
            all_flerrs = cos_spectrum.segments[segment].fluxerrs
            all_dqs = cos_spectrum.segments[segment].dqs
            all_segments = numpy.repeat(segment, n_wls)
        else:
            all_wls = numpy.append(
                all_wls, cos_spectrum.segments[segment].wavelengths)
            n_wls = len(cos_spectrum.segments[segment].wavelengths)
            all_fls = numpy.append(
                all_fls, cos_spectrum.segments[segment].fluxes)
            all_flerrs = numpy.append(
                all_flerrs, cos_spectrum.segments[segment].fluxerrs)
            all_dqs = numpy.append(
                all_dqs, cos_spectrum.segments[segment].dqs)
            all_segments = numpy.append(
                all_segments, numpy.repeat(segment, n_wls))
        # Add this segments suggested y-range to the list so we can add it
        # to the header as keywords later.
        segment_ylims.append({'segment' : segment,
                              'ymin' : plot_metrics[i]['y_axis_range'][0],
                              'ymax' : plot_metrics[i]['y_axis_range'][1]})

    # Mask any wavelengths that are within an Avoid Region.
    all_masks = numpy.repeat(0, len(all_wls))
    for pmetric in plot_metrics:
        for avregion in pmetric['avoid_regions']:
            mask_indices = numpy.where((all_wls >= avregion.minwl) &
                                       (all_wls <= avregion.maxwl))[0]
            all_masks[mask_indices] = 1

    # Output the FITS file.
    wl_col = fits.Column(name='wls', format='E', array=all_wls, unit="angstrom")
    fl_col = fits.Column(name='fls', format='D', array=all_fls,
                         unit="erg/s/cm**2/angstrom")
    fle_col = fits.Column(name='flerrs', format='D', array=all_flerrs,
                          unit="erg/s/cm**2/angstrom")
    dq_col = fits.Column(name='dq_wgts', format='I', array=all_dqs)
    seg_col = fits.Column(name='segment', format='4A', array=all_segments)
    mask_col = fits.Column(name='masked', format='I', array=all_masks)

    # Create the extension HDU.
    cols = fits.ColDefs([wl_col, fl_col, fle_col, dq_col, seg_col, mask_col])
    hdu1 = fits.BinTableHDU.from_columns(cols)

    # Add the suggested ymin and ymax values for each segment.
    for ylim in segment_ylims:
        hdu1.header.set(ylim['segment'].strip().upper()+'YMIN', ylim['ymin'],
                        "Suggested min. for " +
                        ylim['segment'].strip().upper() + " y-axis plot range.")
        hdu1.header.set(ylim['segment'].strip().upper()+'YMAX', ylim['ymax'],
                        "Suggested max. for " +
                        ylim['segment'].strip().upper() + " y-axis plot range.")

    # Create a barebones primary HDU.
    hdr = fits.Header()
    hdr.set('ORIGFILE', os.path.basename(input_file), "Input file name.")
    hdu0 = fits.PrimaryHDU(header=hdr)

    # Output the FITS file.
    hdulist = fits.HDUList([hdu0, hdu1])
    hdulist.writeto(output_file, overwrite=True)
#--------------------



"""
.. module:: avoidregion
   :synopsis: Create AvoidRegions in wavelength space that will not be used
       when determining the optimal y-axis plot rangex.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
from builtins import object
from builtins import str
#--------------------
# Package Imports
#--------------------
from spec_plots import __version__

#--------------------

class AvoidRegion(object):
    """
    Defines an avoid region, which is a section of wavelength space that should
    not be included when determining the optimal y-axis plot range.  The object
    consists of a starting wavelength, ending wavelength, and string description
    of what that region is.

    :raises: ValueError
    """
    def __init__(self, minwl=None, maxwl=None, description=""):

        if minwl is None:
            raise ValueError("Must specify a minimum wavelength for this avoid"
                             " region.")

        if maxwl is None:
            raise ValueError("Must specify a maximum wavelength for this avoid"
                             " region.")

        if minwl >= maxwl:
            raise ValueError("Minimum wavelength must be less than maximum"
                             " wavelength for this avoid region.  Given min."
                             " wavelength = " + str(minwl) + " and max."
                             " wavelength = " + str(maxwl) + ".")

        # Assign the min. wl., max. wl., and description to the object.
        self.minwl = minwl
        self.maxwl = maxwl
        self.description = description
#--------------------


#--------------------
def generate_avoid_regions(instrument):
    """
    Creates a list of AvoidRegion objects for use in the plotting routine,
    specifically designed for the given instrument.

    :param instrument: The instrument to generate the Avoid Region for.

    :type instrument: str

    """

    if instrument == "cos":
        lya1215_ar = AvoidRegion(1214., 1217., "Lyman alpha emission line.")
        return [lya1215_ar]

    elif instrument == "stis":
        lya1215_ar = AvoidRegion(1214., 1217., "Lyman alpha emission line.")
        return [lya1215_ar]

    else:
        return []
#--------------------



"""
.. module:: calc_plot_metrics
   :synopsis: Calculates metrics related to a plot, including flux statistics,
       avoid regions, and optimal x- and y-axis ranges for plots.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
#--------------------
# External Imports
#--------------------
import matplotlib.collections
import numpy
#--------------------
# Package Imports
#--------------------
from spec_plots.utils.specutils.get_flux_stats import get_flux_stats
from spec_plots.utils.specutils.avoidregion import generate_avoid_regions
from spec_plots.utils.specutils.set_plot_xrange import set_plot_xrange
from spec_plots.utils.specutils.set_plot_yrange import set_plot_yrange
from spec_plots import __version__

#--------------------

def calc_plot_metrics(instrument, wls, fls, flerrs, dqs, n_consecutive,
                      flux_scale_factor, fluxerr_scale_factor):
    """
    Calculates a variety of plot metrics, including flux statistics, avoid
    regions, and optimal x- and y-axis ranges.

    :param instrument: The instrument that is being tested.

    :type instrument: str

    :param wls: The wavelengths to be plotted.

    :type wls: numpy.ndarray

    :param fls: The fluxes to be plotted.

    :type fls: numpy.ndarray

    :param flerrs: The uncertainties of the fluxes to be plotted.

    :type flerrs: numpy.ndarray

    :param dqs: The DQ flags of the spectrum to be plotted.  For COS, these are
        the DQ_WGT bits from the header.

    :type dqs: numpy.ndarray

    :param n_consecutive: How many consecutive points must pass the test for
        the index to count as the valid start/end of the spectrum?

    :type n_consecutive: int

    :param flux_scale_factor: Max. allowed ratio between the flux and a median
        flux value, used in edge trimming.

    :type flux_scale_factor: float

    :param fluxerr_scale_factor: Max. allowed ratio between the flux uncertainty
        and a median flux uncertainty value, used in edge trimming.

    :type fluxerr_scale_factor: float

    :returns: dict -- A container for the plot metrics.
    """

    # Calculate statistics on the fluxes.
    median_flux, median_fluxerr, fluxerr_95th = get_flux_stats(fls, flerrs)

    # Determine optimal x-axis.  This is not the x-axis plot range used, but
    # rather the area of the plot that is considered when scaling the y-axis.
    optimal_xaxis_range = set_plot_xrange(instrument, wls, fls, flerrs, dqs,
                                          n_consecutive, flux_scale_factor,
                                          fluxerr_scale_factor, median_flux,
                                          median_fluxerr, fluxerr_95th)

    # Create COS avoid regions.
    avoid_regions = generate_avoid_regions(instrument)

    # Determine the optimal y-axis.
    if all(numpy.isfinite(optimal_xaxis_range)):
        y_axis_range = set_plot_yrange(wls, fls, avoid_regions=avoid_regions,
                                       wl_range=optimal_xaxis_range)
    else:
        y_axis_range = [numpy.nan, numpy.nan]

    # Construct the LineCollection of segments for this curve.   Only do so if
    # the plot transparency will be < 1.0.
    points = numpy.array([wls, fls]).T.reshape(-1, 1, 2)
    segments = numpy.concatenate([points[:-1], points[1:]], axis=1)
    linecoll = matplotlib.collections.LineCollection(segments)

    # Return the plot_metrics dict.
    return {"median_flux":median_flux, "median_fluxerr":median_fluxerr,
            "fluxerr_95th":fluxerr_95th, "optimal_xaxis_range":
                optimal_xaxis_range, "avoid_regions":avoid_regions,
            "y_axis_range":y_axis_range, "line_collection":linecoll}
#--------------------




"""
.. module:: jwstspectrum
   :synopsis: Class definitions for a JWST Spectrum object.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
from builtins import object
#--------------------
# Package Imports
#--------------------
from spec_plots import __version__

#--------------------

class JWSTSpectrum(object):
    """
    Defines a JWST spectrum, including wavelegnth, flux, flux errors, and DQ
    values.

    :raises: ValueError
    """
    def __init__(self, wl_arr, fl_arr, fle_arr, dq_arr, orig_file=None):
        """
        Create a JWSTSpectrum object.

        :param wl_arr: The array of wavelength values.

        :type wl_arr: numpy.ndarray

        :param fl_arr: The array of flux values.

        :type fl_arr: numpy.ndarray

        :param fle_arr: The array of flux uncertainty values.

        :type fle_arr: numpy.ndarray

        :param dq_arr: The array of DQ values.

        :type dq_arr: numpy.ndarray

        :param orig_file: Original FITS file read to create the spectrum
            (includes full path).

        :type orig_file: str
        """

        # Record the original file name.
        self.orig_file = orig_file

        # Record the wavelength, flux, flux uncertainty, and DQ values.
        self.wavelengths = wl_arr
        self.fluxes = fl_arr
        self.fluxerrs = fle_arr
        self.dqs = dq_arr
#--------------------



"""
.. module:: readspec
   :synopsis: Reads in a JWST spectrum from a FITS file.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
from __future__ import print_function
#--------------------
# External Imports
#--------------------
from astropy.io import fits
#--------------------
# Package Imports
#--------------------
from spec_plots import __version__
from spec_plots.utils.specutils_jwst.jwstspectrum import JWSTSpectrum

#--------------------

#--------------------
def readspec(input_file):
    """
    Reads in a JWST spectrum FITS file (x1d, x1dints) and
    returns the wavelengths, fluxes, flux uncertainties, and DQ values.

    :param input_file: Name of input FITS file.

    :type input_file: str

    :returns: JWSTSpectrum -- The spectroscopic data (wavelength, flux,
        flux error, etc):

    :raises: KeyError
    """

    with fits.open(input_file) as hdulist:

        # Read the data from the first extension.  For JWST, the spectra are
        # always stored as tables in the first FITS extension.
        jwst_tabledata = hdulist[1].data

        # Extract wavelength, fluxes, flux uncertainties, and DQ flags for
        # each segment.
        try:
            wavelength_table = jwst_tabledata.field("WAVELENGTH")
        except KeyError:
            print("*** MAKE_JWST_SPEC_PREVIEWS ERROR: WAVELENGTH column not"
                  " found in first extension's binary table.")
            exit(1)

        try:
            flux_table = jwst_tabledata.field("FLUX")
        except KeyError:
            print("*** MAKE_JWST_SPEC_PREVIEWS ERROR: FLUX column not found in"
                  " first extension's binary table.")
            exit(1)

        try:
            fluxerr_table = jwst_tabledata.field("ERROR")
        except KeyError:
            print("*** MAKE_JWST_SPEC_PREVIEWS ERROR: ERROR column not found"
                  " in first extension's binary table.")
            exit(1)

        try:
            dq_table = jwst_tabledata.field("DQ")
        except KeyError:
            print("*** MAKE_JWST_SPEC_PREVIEWS ERROR: DQ column not found"
                  " in first extension's binary table.")
            exit(1)

        # Create JWSTSpectrum object.
        return_spec = JWSTSpectrum(wavelength_table, flux_table, fluxerr_table,
                                   dq_table, orig_file=input_file)

        return return_spec
#--------------------



"""
.. module:: dq_has_flag
   :synopsis: Utility function that returns True if a given Data Quality bitmask
       contains a specific flag.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
from builtins import str
#--------------------
# Package Imports
#--------------------
from spec_plots import __version__

#--------------------

def dq_has_flag(dqf, flag_to_check):
    """
    Returns true/false if the given DQ value has a specific flag value set after
    unpacked into a 16-bit string.  For example:

    .. code-block:: python

         print dq_has_flag(48,16)
         True
         print dq_has_flag(40, 16)
         False

    :param dqf: The DQ value.

    :type dqf: int

    :param flag_to_check: The flag value to check if it's set to True.

    :type flag_to_check: int

    :returns: bool -- Returns True if `flag_to_check` is `True` inside `dqf`.

    :raises: ValueError
    """

    # Make sure `flag_to_check` is a power of 2.
    if (flag_to_check & (flag_to_check-1)) == 0 and flag_to_check != 0:
        dq_16bit_str = "{0:b}".format(dqf)
        flag_16bit_str = "{0:b}".format(flag_to_check)

        # If the 16-bit string of the value to check is longer than 16-bit
        # string version of the DQ value, then we know it can't be part of the
        # DQ  bitmask.  If not, then look for that bit to be set (by counting
        # from the right).
        return (len(flag_16bit_str) <= len(dq_16bit_str) and
                dq_16bit_str[-1 * len(flag_16bit_str)] == '1')
    else:
        raise ValueError("Flag to check must be a power of 2.  Asked to check"
                         " whether flag " + str(flag_to_check) + " is set to"
                         " True in bitmask value " + str(dqf) + ", but " +
                         str(flag_to_check) + " is not a power of 2.")
#--------------------



"""
.. module:: __init__

   :synopsis: Used to treat "specutils" as a directory and package.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
#--------------------
# Package Imports
#--------------------
from spec_plots import __version__
from spec_plots.utils.specutils.specutilserror import SpecUtilsError
from spec_plots.utils.specutils.avoidregion import (AvoidRegion,
                                                    generate_avoid_regions)
from spec_plots.utils.specutils.is_bad_dq import is_bad_dq
from spec_plots.utils.specutils.stitch_components import stitch_components
from spec_plots.utils.specutils.calc_covering_fraction import (
    calc_covering_fraction)
from spec_plots.utils.specutils.calc_plot_metrics import calc_plot_metrics
from spec_plots.utils.specutils.debug_oplot import debug_oplot
from spec_plots.utils.specutils.dq_has_flag import dq_has_flag
from spec_plots.utils.specutils.edge_trim import (edge_trim,
                                                  _set_plot_xrange_test)
from spec_plots.utils.specutils.get_flux_stats import get_flux_stats
from spec_plots.utils.specutils.set_plot_xrange import set_plot_xrange
from spec_plots.utils.specutils.set_plot_yrange import set_plot_yrange



"""
.. module:: get_segment_names
   :synopsis: Returns a list of segment names such that the bluest segment
       comes first.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
#--------------------
# Package Imports
#--------------------
from spec_plots import __version__

#--------------------

def get_segment_names(cos_spectrum):
    """
    Returns a list of segment names sorted such that the bluest segment come
    first.

    :param cos_spectrum: COS spectrum as returned by READSPEC.

    :type cos_spectrum: COSSpectrum

    :returns: list -- A list of segment names.
    """

    # Get an initial list of segment names.
    segment_names = list(cos_spectrum.segments.keys())

    # Reverse the list of segment names *FOR FUV DATA*, becaue the bluest
    # segment is the latter in the alphabet, but only for the FUV spectra.  If
    # NUV, then just make sure the segments are sorted alphabetically.
    if cos_spectrum.band == 'FUV':
        segment_names.sort(reverse=True)
    else:
        segment_names.sort(reverse=False)

    return segment_names
#--------------------



#!/usr/bin/env python

"""
.. module:: make_hst_spec_previews
   :synopsis: Given an HST spectrum file name, this script will read in the
       data and generate preview plots of the spectra.  The plots are generated
       in different dimensions (large, thumbnail) and, depending on the
       instrument / configuration, are plotted in different formats to maximize
       readability and usability.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
from __future__ import print_function
import argparse
from os import path
from builtins import str
#--------------------
# External Imports
#--------------------
from astropy.io import fits
import numpy
#--------------------
# Package Imports
#--------------------
from spec_plots.utils import specutils, specutils_cos, specutils_stis
from spec_plots import __version__

FLUX_SCALE_FACTOR_DEFAULT = 10.
FLUXERR_SCALE_FACTOR_DEFAULT = 5.
N_CONSECUTIVE_DEFAULT = 20
OUTPUT_PATH_DEFAULT = ""
OUTPUT_TYPE_DEFAULT = [str("png")]
DPI_VAL_DEFAULT = 96.
DEBUG_DEFAULT = False
FULL_YLABELS_DEFAULT = False
NOOPTIMIZE_DEFAULT = False
VERBOSE_DEFAULT = False

#--------------------

class HSTSpecPrevError(Exception, object):
    """
    This class defines a generic Exception to use for errors raised in
    MAKE_HST_SPEC_PREVIEWS.  It simply prints the given string when raising the
    exception. e.g.,

    .. code-block:: python

         raise HSTSpecPrevError("Print this string")
         HSTSpecPrevError: *** MAKE_HST_SPEC_PREVIEWS ERROR: 'Print this string'
    """

    def __init__(self, value):
        """
        Initiate the Exception.

        :param value: The string to print on error.

        :type value: str
        """
        super(HSTSpecPrevError, self).__init__(value)
        self.value = value

    def __str__(self):
        """
        Overrides the str function for this class.
        """
        return "*** MAKE_HST_SPEC_PREVIEWS ERROR: "+repr(self.value)

#--------------------

def check_input_options(args):
    """
    Check that input arguments satisfy some minimum requirements.

    :param args: Stores arguments and options.

    :type args: argparse.Namespace object.

    :raises: HSTSpecPrevError, ValueError
    """

    # Make sure the input file is trimmed for use later on in the program.
    args.input_file = args.input_file.strip()

    # Make sure the output_type string(s) is(are) trimmed and lowercase.
    args.output_type = [x.strip().lower() for x in args.output_type]

    # The DPI value must be greater than zero...
    if args.dpi_val <= 0.:
        raise ValueError("DPI value must be > 0.")

#--------------------

def get_instrument_name(input_file):
    """
    Retrieves the instrument name from a FITS file primary header.

    :param input_file: Name of input FITS file.

    :type input_file: str

    :returns: str -- The name of the instrument based on the FITS header
        keyword, in uppercase with leading/trailing whitespace removed.

    :raises: KeyError, HSTSpecPrevError
    """

    # Make sure file exists.
    if not path.isfile(input_file):
        raise HSTSpecPrevError("Input file not found, looking for file: " +
                               input_file)

    with fits.open(input_file) as hdulist:
        # Make sure the INSTRUME keyword exists in the primary header,
        # otherwise, catch as a KeyError in this case.
        try:
            this_instrument = hdulist[0].header["INSTRUME"]
        except KeyError:
            print("*** MAKE_HST_SPEC_PREVIEWS ERROR: INSTRUME keyword not"
                  " found in this file's primary header: " + input_file)
            exit(1)
    return this_instrument.strip().upper()

#--------------------

def make_hst_spec_previews(input_file, flux_scale_factor=
                           FLUX_SCALE_FACTOR_DEFAULT, fluxerr_scale_factor=
                           FLUXERR_SCALE_FACTOR_DEFAULT, n_consecutive=
                           N_CONSECUTIVE_DEFAULT, output_path=
                           OUTPUT_PATH_DEFAULT, output_type=OUTPUT_TYPE_DEFAULT,
                           dpi_val=DPI_VAL_DEFAULT, debug=DEBUG_DEFAULT,
                           full_ylabels=FULL_YLABELS_DEFAULT, optimize=
                           not NOOPTIMIZE_DEFAULT, verbose=VERBOSE_DEFAULT):
    """
    Main function in the module.

    :param input_file: The full path and name of the FITS file to create a
        preview for.

    :type input_file: str

    :param flux_scale_factor: The ratio between the flux and the median flux
        that defines the pass/fail criterion within the edge trim test.

    :type flux_scale_factor: float

    :param fluxerr_scale_factor: The ratio between the flux uncertainty and the
        median flux uncertainty that defines the pass/fail criterion within the
        edge trim test.

    :type fluxerr_scale_factor: float

    :param n_consecutive: The number of consecutive data points that must pass
        the edge trim test to define the start and end of the spectrum for
        plotting purposes.

    :type n_consecutive: int

    :param output_path: Full path to output plot files.  Do not include the
        output file name in this path.  If not supplied, plots will be created
        in the same directory as the input file.

    :type output_path: str

    :param output_type: The file type(s) of the plots to make.

    :type output_type: list

    :param dpi_val: The DPI value of your device's monitor, which will affect
        the size of the output plots.

    :type dpi_val: float

    :param debug: If True, turns on debug mode, which will plot to the screen
        and color-code fluxes based on different rejection criteria.

    :type debug: bool

    :param full_ylabels: If True, label the y-axis with full values, including
        powers of ten in scientific notation.

    :type full_ylabels: bool

    :param optimize: If set to True, will use a slightly optimized version of
        determining the plot covering fraction.

    :type optimize: bool

    :param verbose: Turn on verbose messages/logging?

    :type verbose: bool

    :raises: HSTSpecPrevError
    """

    # Print file name, if verbose is turned on.
    if verbose:
        print("Input file: " + input_file)

    # Derive output file names from input file name.
    output_files = []
    for out_type in output_type:
        if out_type != "screen":
            if out_type != "fits":
                output_file = (path.join(output_path, "") +
                               path.basename(input_file).split(".fits")[0] +
                               "." + out_type)
            else:
                output_file = (path.join(output_path, "") +
                               path.basename(input_file).split(".fits")[0] +
                               "_prev." + out_type)
        else:
            output_file = None

        output_files.append(output_file)

    # Print name of output file.
    if verbose:
        print("Output file names are:")
        for ofile in output_files:
            if ofile is not None:
                if ofile[-4:] == '.png':
                    print("  Output file: " + ofile)
                    print("  Output file: " + ofile.strip('\.png') +
                          '_thumb.png')
                else:
                    print("  Output file: " + ofile)
            else:
                print("  Plotting to screen.")

    # Read in the FITS file to determine which instrument it comes from.
    # Print the name of the instrument found in the header if verbose is turned
    # on.
    this_instrument = get_instrument_name(input_file)
    if verbose:
        print("Instrument: " + this_instrument)

    # Read in the FITS files and create plots using the local package
    # appropriate for the instrument used in the input file.
    if this_instrument == "COS":
        # Get wavelengths, fluxes, flux uncertainties.
        cos_spectrum = specutils_cos.readspec(input_file)

        # Get a list of segment names sorted such that the bluest segment is
        # first.
        cos_segment_names = specutils_cos.get_segment_names(cos_spectrum)

        # Trim the wavelengths < 900 Angstroms for FUVB segment if optical
        # element used is G140L.
        if ("FUVB" in cos_spectrum.segments and cos_spectrum.optical_element ==
                "G140L"):
            specutils_cos.extract_subspec(cos_spectrum, "FUVB", min_wl=900.)

        # Create a stitched spectrum for use when making thumb-size plots.
        stitched_spectrum = specutils.stitch_components(cos_spectrum,
                                                        n_consecutive,
                                                        flux_scale_factor,
                                                        fluxerr_scale_factor,
                                                        segment_names=
                                                        cos_segment_names)

        # Calculate plot metrics for the each segment.
        segment_plot_metrics = [
            specutils.calc_plot_metrics("cos",
                                        cos_spectrum.segments[x].wavelengths,
                                        cos_spectrum.segments[x].fluxes,
                                        cos_spectrum.segments[x].fluxerrs,
                                        cos_spectrum.segments[x].dqs,
                                        n_consecutive, flux_scale_factor,
                                        fluxerr_scale_factor)
            for x in cos_segment_names]

        # Make "large-size" plot.
        for out_type, out_file in zip(output_type, output_files):
            if out_type != "fits":
                specutils_cos.plotspec(cos_spectrum, out_type, out_file,
                                       flux_scale_factor,
                                       fluxerr_scale_factor,
                                       segment_plot_metrics,
                                       dpi_val=dpi_val, output_size=1024,
                                       debug=debug, full_ylabels=full_ylabels,
                                       title_addendum=stitched_spectrum["title"],
                                       optimize=optimize)
            else:
                # Write the spectrum that would be plotted to a binary FITS
                # table.
                specutils_cos.make_fits(cos_spectrum, out_file,
                                        segment_plot_metrics, input_file)

        if not debug and output_type != ["fits"]:
            # Calculate plot metrics for the stitched spectrum.
            stitched_plot_metrics = [
                specutils.calc_plot_metrics("cos", stitched_spectrum["wls"],
                                            stitched_spectrum["fls"],
                                            stitched_spectrum["flerrs"],
                                            stitched_spectrum["dqs"],
                                            n_consecutive, flux_scale_factor,
                                            fluxerr_scale_factor)]

            # Make "thumbnail-size" plot, if requested.
            for out_type, out_file in zip(output_type, output_files):
                if out_type != 'fits':
                    specutils_cos.plotspec(cos_spectrum, out_type,
                                           out_file, flux_scale_factor,
                                           fluxerr_scale_factor,
                                           stitched_plot_metrics,
                                           dpi_val=dpi_val, output_size=128,
                                           stitched_spectrum=stitched_spectrum,
                                           optimize=optimize)

    elif this_instrument == "STIS":
        # Get wavelengths, fluxes, flux uncertainties.
        stis_spectrum = specutils_stis.readspec(input_file)

        # Get the indices to plot.  If the number of associations is <=3 then
        # we will plot all of them, but if >3 then only the first, middle, and
        # last association will be plotted, so it's not necessary to stitch or
        # calculate plot metrics for any of the others.
        indices_to_plot = specutils_stis.get_association_indices(
            stis_spectrum.associations)

        # Create a stitched spectrum for each association, used when making
        # thumb-size plots.  The stitched spectrum joins the orders together (if
        # there is more than one).  The length of the returned list is equal to
        # the number of associations that will be plotted.
        stitched_spectra = [
            specutils.stitch_components(x, n_consecutive, flux_scale_factor,
                                        fluxerr_scale_factor)
            for x in numpy.asarray(stis_spectrum.associations)[indices_to_plot]]

        # Calculate plot metrics for the each association that will be
        # plotted.  The length of the returned list is equal to the number of
        # associations that will be plotted.
        association_plot_metrics = [
            specutils.calc_plot_metrics("stis", x["wls"], x["fls"], x["flerrs"],
                                        x["dqs"], n_consecutive,
                                        flux_scale_factor, fluxerr_scale_factor)
            for x in stitched_spectra]

        # Make "large-size" plot.
        for out_type, out_file in zip(output_type, output_files):
            if out_type != 'fits':
                specutils_stis.plotspec(stis_spectrum, indices_to_plot,
                                        stitched_spectra, out_type, out_file,
                                        flux_scale_factor, fluxerr_scale_factor,
                                        association_plot_metrics,
                                        dpi_val=dpi_val, output_size=1024,
                                        debug=debug, full_ylabels=full_ylabels,
                                        optimize=optimize)

                # Make "thumbnail-size" plot, if requested.  Notice that in this
                # case we always plot just the first association, by passing only
                # `stitched_spectra[0]`.
                if not debug:
                    specutils_stis.plotspec(stis_spectrum, indices_to_plot,
                                            [stitched_spectra[0]], out_type,
                                            out_file, flux_scale_factor,
                                            fluxerr_scale_factor,
                                            association_plot_metrics,
                                            dpi_val=dpi_val,
                                            output_size=128, optimize=optimize)

    else:
        raise HSTSpecPrevError("'INSTRUME' keyword not understood: " +
                               this_instrument)

#--------------------

def setup_args():
    """
    Set up command-line arguments and options.

    :returns: ArgumentParser -- Stores arguments and options.
    """
    parser = argparse.ArgumentParser(
        description="Create spectroscopic preview plots given an HST spectrum"
        " FITS file.")

    parser.add_argument("input_file", action="store", type=str, help=
                        "[Required] Full path to input file (HST spectrum) for"
                        " which to generate preview plots.  Include the file"
                        " name in the path.")

    parser.add_argument("-d", action="store_true", dest="debug", default=
                        DEBUG_DEFAULT, help="[Optional] Turn on debug mode,"
                        " which will plot to the screen and color-code fluxes"
                        " based on different rejection criteria.  Default ="
                        " %(default)s.")

    parser.add_argument("--dpival", action="store", type=float, dest="dpi_val",
                        default=DPI_VAL_DEFAULT, help="[Optional] Specify the"
                        " DPI value of your device's monitor, which will affect"
                        " the size of the output plots.  Default ="
                        " %(default)s.")

    parser.add_argument("-e", action="store", type=float, dest=
                        "fluxerr_scale_factor", default=
                        FLUXERR_SCALE_FACTOR_DEFAULT, help="[Optional] Specify"
                        " the ratio between the flux uncertainty and the median"
                        " flux uncertainty that defines the pass/fail criterion"
                        " within the edge trim test.  Default = %(default)s.")

    parser.add_argument("-n", action="store", type=int, dest="n_consecutive",
                        default=N_CONSECUTIVE_DEFAULT, help="[Optional] Specify"
                        " the number of consecutive data points that must pass"
                        " the edge trim test to define the start and end of the"
                        " spectrum for plotting purposes.  Default ="
                        " %(default)i.")

    parser.add_argument("--nooptimize", action="store_true", dest="nooptimize",
                        default=NOOPTIMIZE_DEFAULT, help="[Optional] If True,"
                        " do not use an optimized method of calculating plot"
                        " covering fraction to determine line transparency."
                        "  Default = %(default)s.")

    parser.add_argument("-o", action="store", type=str, dest="output_path",
                        default=OUTPUT_PATH_DEFAULT, help="[Optional] Full path"
                        " to output plot files.  Do not include file name in"
                        " path.    Default is to output to the same directory"
                        " as the input file.")

    parser.add_argument("-s", action="store", type=float, dest=
                        "flux_scale_factor", default=FLUX_SCALE_FACTOR_DEFAULT,
                        help="[Optional] Specify the ratio between the flux and"
                        " the median flux that defines the pass/fail criterion"
                        " within the edge trim test.  Default = %(default)s.")

    parser.add_argument("-t", nargs='+', action="store", type=str, dest=
                        "output_type", default=OUTPUT_TYPE_DEFAULT, help=
                        "[Optional] Specify the file type of the plots to make."
                        "  Default = %(default)s.", choices=['png', 'eps',
                                                             'screen', 'fits'],
                        metavar="{png,ps,screen,fits}")

    parser.add_argument("-v", action="store_true", dest="verbose",
                        default=VERBOSE_DEFAULT, help="[Optional] Turn on"
                        " verbose messages/logging?  Default = %(default)s.")

    parser.add_argument("-y", action="store_true", dest="full_ylabels",
                        default=FULL_YLABELS_DEFAULT, help="[Optional] Label"
                        " the y-axis with full values, including powers of ten"
                        " in scientific notation?  Default = %(default)s.")

    return parser

#--------------------

if __name__ == "__main__":

    # Create ArgumentParser object that holds arguments and options.
    INPUT_ARGS = setup_args().parse_args()

    # Check arguments and options.
    check_input_options(INPUT_ARGS)

    # Call main function.
    make_hst_spec_previews(INPUT_ARGS.input_file,
                           flux_scale_factor=INPUT_ARGS.flux_scale_factor,
                           fluxerr_scale_factor=INPUT_ARGS.fluxerr_scale_factor,
                           n_consecutive=INPUT_ARGS.n_consecutive,
                           output_path=INPUT_ARGS.output_path,
                           output_type=INPUT_ARGS.output_type,
                           dpi_val=INPUT_ARGS.dpi_val,
                           debug=INPUT_ARGS.debug,
                           full_ylabels=INPUT_ARGS.full_ylabels,
                           optimize=not INPUT_ARGS.nooptimize,
                           verbose=INPUT_ARGS.verbose)
#--------------------



"""
.. module:: set_plot_xrange
   :synopsis: Determines the optimal x-axis plot range, in wavelength space.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
from __future__ import print_function
#--------------------
# External Imports
#--------------------
import numpy
#--------------------
# Package Imports
#--------------------
from spec_plots.utils.specutils.edge_trim import edge_trim
from spec_plots import __version__

#--------------------

def set_plot_xrange(instrument, wavelengths, fluxes, fluxerrs, dqs,
                    n_consecutive, flux_scale_factor, fluxerr_scale_factor,
                    median_flux, median_fluxerr, fluxerr_95th):
    """
    Given an array of wavelengths and fluxes, returns a list of [xmin,xmax] to
    define an optimal x-axis plot range.

    :param instrument: The instrument that is being tested.

    :type instrument: str

    :param wavelengths: The wavelengths to be plotted.

    :type wavelengths: numpy.ndarray

    :param fluxes: The fluxes to be plotted.

    :type fluxes: numpy.ndarray

    :param fluxerrs: The uncertainties of the fluxes to be plotted.

    :type fluxerrs: numpy.ndarray

    :param dqs: The DQ flags of the spectrum to be plotted.  For COS, these are
        the DQ_WGT bits from the header.

    :type dqs: numpy.ndarray

    :param n_consecutive: How many consecutive points must pass the test for the
        index to count as the valid start/end of the spectrum?

    :type n_consecutive: int

    :param flux_scale_factor: Max. allowed ratio between the flux and a median
        flux value, used in edge trimming.

    :type flux_scale_factor: float

    :param fluxerr_scale_factor: Max. allowed ratio between the flux uncertainty
        and a median flux uncertainty value, used in edge trimming.

    :type fluxerr_scale_factor: float

    :param median_flux: A median flux value, used in the test.

    :type median_flux: float

    :param median_fluxerr: A median flux uncertainty value, used in the test.

    :type median_fluxerr: float

    :param fluxerr_95th: The flux uncertainty corresponding to the 95th
        percentile.

    :type fluxerr_95th: float

    :returns: list -- Two-element list containing the optimal [xmin,xmax] values
        to define the x-axis plot range.
    """

    # Test if there are any NaN's in the wavelength array.  If so, issue a
    # warning for now...
    if numpy.isnan(numpy.sum(wavelengths)):
        print ("***WARNING in SPECUTILS_STIS: Wavelength array contains NaN"
               " values.  Behavior has not been fully tested in this case.")

    # Is the array of fluxes all zero values?  If so, then immediately return
    # NaN values for the edges.
    if len(numpy.unique(fluxes)) == 1 and fluxes[0] == 0.:
        return [numpy.nan, numpy.nan]

    # Find the first element in the array that is NOT considered an "edge
    # effect", and the last element in the array that is NOT considered an "edge
    # effect".  If the input array is all zeroes, then it will find the last and
    # first element, respectively.  Note that the trim does not just stop at the
    # first index that satisfies this requirement, since there can be spikes at
    # the edges that can fool the algorithm.  Instead, it requires that the next
    # "n_consecutive" data points after each trial location also fail the test
    # for "edge effect".  It returns start and stop indexes taking into account
    # DQ values, as well as start and stop indexes without taking into account
    # DQ values.
    start_index_nodq, end_index_nodq, start_index_withdq, end_index_withdq = (
        edge_trim(instrument, fluxes, fluxerrs, dqs, n_consecutive, median_flux,
                  flux_scale_factor, median_fluxerr, fluxerr_scale_factor,
                  fluxerr_95th))

    # Return the optimal start and end wavelength values for defining the
    # x-axis plot range.  Note that if the fluxes are all zeroes, then start
    # index will be past end index, so we return NaN values to indicate a
    # special plot should be made in that case.  The odd conditional below
    # checks to make sure the end index (working from the back of the list via
    # negative indexes) stops before reaching the start index (which works from
    # the front using zero-based, positive indexes), otherwise return NaN values
    # because the array is all zeroes.
    if len(fluxes) + end_index_withdq > start_index_withdq:
        return [wavelengths[start_index_withdq], wavelengths[end_index_withdq]]
    elif len(fluxes) + end_index_nodq > start_index_nodq:
        return [wavelengths[start_index_nodq], wavelengths[end_index_nodq]]
    else:
        return [numpy.nan, numpy.nan]
#--------------------



"""
.. module:: extract_subspec
   :synopsis: Extracts a sub-spectrum from the provided COS spectrum object.
       Modifies the COS spectrum object to replace that segment's spectrum with
       the specified sub-spectrum.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
from __future__ import print_function
import os
import sys
from builtins import str
#--------------------
# External Imports
#--------------------
import numpy
#--------------------
# Package Imports
#--------------------
from spec_plots.utils.specutils.specutilserror import SpecUtilsError
from spec_plots import __version__
#--------------------

#--------------------
def extract_subspec(cos_spectrum, segment, min_wl=None, max_wl=None):
    """
    Extracts a sub-spectrum of a COS segment's spectrum.  If both min. and max.
    wavelengths are specified, then it keeps the part of the spectrum within
    those bounds.  If only a min. or max. is specified, then those are treated
    as lower or upper bounds, and the part of the spectrum redder/bluer than the
    bound is retained.

    :param cos_spectrum: COS spectrum as returned by READSPEC.

    :type cos_spectrum: COSSpectrum

    :param segment: The segment to do the extraction on.

    :type segment: str

    :param min_wl: The minimum wavlength value to keep.  If None, then there
        will be no lower bound.

    :type min_wl: float

    :param max_wl: The maximum wavelength value to keep.  If None, then there
        will be no upper bound.

    :type max_wl: float
    """

    if segment in cos_spectrum.segments:
        if min_wl is not None and max_wl is not None:
            where_within = numpy.where(
                (cos_spectrum.segments[segment].wavelengths >= min_wl) &
                (cos_spectrum.segments[segment].wavelengths <= max_wl))[0]
        elif min_wl is not None and max_wl is None:
            where_within = numpy.where(
                (cos_spectrum.segments[segment].wavelengths >= min_wl))[0]
        elif min_wl is None and max_wl is not None:
            where_within = numpy.where(
                (cos_spectrum.segments[segment].wavelengths <= max_wl))[0]

        n_within = len(where_within)
        if n_within > 0:
            # Then we extract the subspectrum for this object by modifying the
            # cos_spectrum object.
            cos_spectrum.segments[segment].nelem = n_within
            cos_spectrum.segments[segment].wavelengths = (
                cos_spectrum.segments[segment].wavelengths[where_within])
            cos_spectrum.segments[segment].fluxes = (
                cos_spectrum.segments[segment].fluxes[where_within])
            cos_spectrum.segments[segment].fluxerrs = (
                cos_spectrum.segments[segment].fluxerrs[where_within])
            cos_spectrum.segments[segment].dqs = (
                cos_spectrum.segments[segment].dqs[where_within])
        else:
            print("*** WARNING in SPECUTILS_COS: Requested subspectrum does"
                  " not overlap with this segment's spectrum.  No extraction"
                  " will be done.")
    else:
        raise SpecUtilsError("The segment where you want to perform the"
                             " subspectrum extraction is not present."
                             "  Specified \"" + segment +
                             "\", available segments for this spectrum are: " +
                             ', '.join(cos_spectrum.segments)+".")
#--------------------



"""
.. module:: __init__

   :synopsis: Used to treat "specutils_jwst" as a directory and package.

.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""

#--------------------
# Built-In Imports
#--------------------
from __future__ import absolute_import
#--------------------
# Package Imports
#--------------------
from spec_plots import __version__
from spec_plots.utils.specutils_jwst.plotspec import plotspec
from spec_plots.utils.specutils_jwst.readspec import readspec
